// Program from Stephan Mock 324612 and Randolf Schrfig 327639
// modifiziert von cg.in.tu-clausthal

#include <limits>
#include <QtOpenGL/QGLWidget>

#include "mesh.h"

/* Draw the mesh using smooth or flat normals accordingly to the parameter */
void Mesh::draw() 
{
	const bool normalsAvailable = (0 < m_normals.size());
	bool texcoordsAvailable = false;
	//glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
	if( normalsAvailable )
	{
		glEnable(GL_LIGHTING);
	} else {
		glDisable(GL_LIGHTING);
	}

	for(unsigned int mesh_Counter = 0; mesh_Counter < m_meshPartInfo.size(); mesh_Counter++)
	{
		glDisable(GL_TEXTURE_2D);	
		// assign material
		int matidx = m_meshPartInfo[mesh_Counter]->materialIndex;
		if( matidx >= 0 ) // if materials available
		{ 
			texcoordsAvailable = (m_texCoords.size() > 0);
			glMaterialfv(GL_FRONT, GL_AMBIENT, m_materialList[matidx]->ambient);
			glMaterialfv(GL_FRONT, GL_DIFFUSE, m_materialList[matidx]->diffuse);
			glMaterialfv(GL_FRONT, GL_SPECULAR, m_materialList[matidx]->specular);
			glMaterialfv(GL_FRONT, GL_SHININESS, &(m_materialList[matidx]->shininess));
			// bind texture
			if( m_materialList[matidx]->textureIndex >=0 )
			{
				int texidx = m_materialList[matidx]->textureIndex;
				glEnable(GL_TEXTURE_2D);
				int texid = m_textures[texidx].textureId;
				glBindTexture(GL_TEXTURE_2D,m_textures[texidx].textureId);
			}
		} 
		else 
		{
			glColor3f(0.1f, 0.1f, 0.1f);
		}
		int from = m_meshPartInfo[mesh_Counter]->fromFace;
		int to = m_meshPartInfo[mesh_Counter]->toFace;
		glBegin( GL_TRIANGLES );
		for (int face_Counter = from; face_Counter < to; face_Counter++) 
		{
			Triangle &triangle = m_triangles[face_Counter];
			for ( int j = 0; j < 3; j++)
			{
				Vector3D &vertex = m_vertices[triangle.v[j]]; 
				if(texcoordsAvailable) 
				{
					Vector3D &texcoord = m_texCoords[triangle.t[j]];
					glTexCoord2f(texcoord.x,texcoord.y);
				}
				if(normalsAvailable)
				{
					Vector3D &normal = m_normals[triangle.n[j]];
					glNormal3f(normal.x, normal.y, normal.z);
				}
				glVertex3f(vertex.x, vertex.y, vertex.z);
			}
		}
		glEnd();	
	}
}  


void Mesh::clear() 
{
	m_vertices.clear();
	m_normals.clear();
	m_texCoords.clear();
	m_triangles.clear();
	m_textures.clear();

	for(unsigned int i = 0; i < m_meshPartInfo.size(); i++)
		delete m_meshPartInfo[i];
	for(unsigned int i = 0; i < m_materialList.size(); i++)
		delete m_materialList[i];

	m_meshPartInfo.clear();
	m_materialList.clear();

	m_bBoxMin = Vector3D( std::numeric_limits<float>::max() );
	m_bBoxMax = Vector3D( -std::numeric_limits<float>::max() );
}


void Mesh::loadMaterials( const char* sFilename)
{

	char ac_line[256];
	char texname[256];
	std::ifstream istr_in( sFilename );
	if ( !istr_in.is_open() )
	{
		std::cerr << "could not read " << sFilename << std::endl;
	}

	MaterialNode *matnode = NULL;

	std::cerr << " reading Materials " << sFilename << std::endl;	
	while( istr_in && ( !istr_in.eof() ) && istr_in.getline( ac_line, 256 ) )
	{
		float   ar, ag, ab, sr, sg, sb, rr, dr, dg, db;
		float   refractiveIndex, shininess;
		// comment
		if ( ac_line[0] == '#' ) continue;

		//Load Material File
		if ( strncmp(ac_line, "newmtl ", 7 ) == 0 )
		{
			char matname[1024];
			if ( sscanf( ac_line, "newmtl %s", matname ) )
			{
				//Put Last material to the MaterialList
				if( 0 != matnode )
				{
					m_materialList.push_back( matnode );
				}

				//Create new Shader
				matnode = new MaterialNode;
				strcpy( matnode->m_name, matname );
			}
		}
		else if ( strncmp(ac_line, "Ns ", 3 ) == 0 )
		{ 
			sscanf( ac_line, "Ns %f", &shininess );
			if ( matnode != NULL )
				matnode->shininess = shininess;
		}
		else if ( strncmp(ac_line, "Ka ", 3 ) == 0 )
		{ 
			sscanf( ac_line, "Ka %f %f %f", &ar, &ag, &ab );
			if ( matnode != NULL )
				matnode->setAmbientColor( ar, ag, ab );
		}
		else if ( strncmp(ac_line, "Kd ", 3 ) == 0 )
		{ 
			sscanf( ac_line, "Kd %f %f %f", &dr, &dg, &db );
			if ( matnode != NULL )
				matnode->setDiffuseColor(  dr, dg, db );
		}
		else if ( strncmp(ac_line, "Ks ", 3 ) == 0 )
		{ 
			sscanf( ac_line, "Ks %f %f %f", &sr, &sg, &sb );
			if ( matnode != NULL )
				matnode->setSpecularColor( sr, sg, sb );
		}
		else if ( strncmp(ac_line, "Ni ", 3 ) == 0 )
		{ 
			sscanf( ac_line, "Ni %f", &refractiveIndex );
		}
		else if ( strncmp(ac_line, "map_Kd ", 7 ) == 0 )
		{ 
			sscanf( ac_line, "map_Kd %s", texname );
			char textureFilename[256];
			replaceFilename(textureFilename,sFilename,texname);
			matnode->textureIndex = getTextureNode(textureFilename);
		}

		else if ( ( strncmp(ac_line, "Tr ", 3 ) == 0 ) )
		{ 
			//Warning: We simply use transparency for refractive color
			//if refractive index != 1.0 && no transparency => set refractive color to 1.0
			sscanf( ac_line, "Tr %f", &rr );
			if ( matnode != NULL )
				matnode->alpha = rr;
		}
		//Alternative version of alpha value
		else if ( ( strncmp(ac_line, "d ", 2 ) == 0 ) )
		{ 
			//Warning: We simply use transparency for refractive color
			sscanf( ac_line, "d %f", &rr );
			if ( matnode != NULL )
				matnode->alpha = rr;
		}
	}

	//Not to forget to add the last Shader to the MaterialList
	if( matnode != NULL )
		m_materialList.push_back( matnode );
}

int Mesh::getTextureNode(char *filename)
{
	int idx = -1;
	for(unsigned int tc = 0; tc < m_textures.size(); tc++)
	{
		if( !strcmp(m_textures[tc].texname,filename) )
		{
			idx = tc;
		}
	}
	if( idx < 0 ) // not yet known texture image
	{
		TextureNode texnode;
		texnode.qImageTexture = 0;
		texnode.textureId = 0;
		strcpy(texnode.texname,filename);
		idx = static_cast<int>(m_textures.size());
		m_textures.push_back(texnode);
	}
	return idx;
}

void Mesh::readObj( const char* sFilename)
{
	this->clear();

	char ac_line[256];

	std::ifstream istr_in( sFilename );
	if ( !istr_in.is_open() )
	{
		std::cerr << "could not read " << sFilename << std::endl;
		m_meshRead = false;
	}
	m_meshRead = true;
	unsigned int vertexcounter = 0;
	unsigned int texcoordcounter = 0;
	unsigned int facecounter = 0;
	std::cerr << " reading " << sFilename << std::endl;	
	while( istr_in && ( !istr_in.eof() ) && istr_in.getline( ac_line, 256 ) )
	{
		// comment
		if ( ac_line[0] == '#' ) continue;

		// vertex
		else if ( strncmp(ac_line, "v ", 2 ) == 0 )
		{
			vertexcounter++;
		}
		else if ( strncmp(ac_line, "vt ", 2 ) == 0 )
		{
			texcoordcounter++;
		}
		// face
		else if ( strncmp( ac_line, "f ", 2) == 0)
		{
			int	i3VI[4] = {0, 0, 0, 0};
			int	i3TI[4] = {0, 0, 0, 0};
			int	i3NI[4] = {0, 0, 0, 0};
			//Quadrangel: add 2 triangles
			if ( 12 == sscanf( ac_line, "f %d//%d//%d %d//%d//%d %d//%d//%d %d//%d//%d", &i3VI[0], &i3TI[0], &i3NI[0], 
				&i3VI[1], &i3TI[1], &i3NI[1], 
				&i3VI[2], &i3TI[2], &i3NI[2],
				&i3VI[3], &i3TI[3], &i3NI[3] ) )
			{			
				facecounter++;					
			}
			else if ( 8 == sscanf( ac_line, "f %d//%d %d//%d %d//%d %d//%d", &i3VI[0], &i3NI[0], 
				&i3VI[1], &i3NI[1], 
				&i3VI[2], &i3NI[2],
				&i3VI[3], &i3NI[3] ) )
			{						
				facecounter++;		
			}
			else if ( 8 == sscanf( ac_line, "f %d/%d %d/%d %d/%d %d/%d", &i3VI[0], &i3NI[0], 
				&i3VI[1], &i3NI[1], 
				&i3VI[2], &i3NI[2],
				&i3VI[3], &i3NI[3] ) )
			{								
				facecounter++;
			}
			else if ( 4 == sscanf( ac_line, "f %d %d %d %d", &i3VI[0], &i3VI[1], &i3VI[2], &i3VI[3] ) )
			{
				facecounter++;
			}

			facecounter++;
		}
	}

	istr_in.close();


	unsigned int ui_normals_created = 0;
	unsigned int actFace = 0;

	std::list<MaterialNode> materialNodes;
	materialNodes.clear();

	m_vertices.reserve( vertexcounter );
	m_normals.reserve( vertexcounter );
	m_texCoords.reserve( texcoordcounter );
	m_triangles.reserve( facecounter );

	std::ifstream istr2_in( sFilename );

	MeshPartInfo *mpi = 0;

	while( istr2_in && ( !istr2_in.eof() ) && istr2_in.getline( ac_line, 256 ) )
	{

		// comment
		if ( ac_line[0] == '#' ) continue;

		//Load Material File
		if ( strncmp(ac_line, "mtllib ", 7 ) == 0 )
		{
			char filename[1024];
			if ( sscanf( ac_line, "mtllib %s", filename ) )
			{
				char newname[256];
				replaceFilename(newname,sFilename,filename);
				loadMaterials( newname);
			}
		}	

		//New Material => push New Material to material List 
		if ( strncmp(ac_line, "usemtl ", 7 ) == 0 )
		{
			if ( mpi != NULL )
			{
				mpi->toFace = actFace;
				m_meshPartInfo.push_back(mpi);
			}

			mpi = new MeshPartInfo();
			mpi->fromFace = actFace;


			bool foundmaterial = false;

			char matname[1024];
			if ( sscanf( ac_line, "usemtl %s", matname ) )
			{
				for( unsigned int m_Counter = 0; m_Counter < m_materialList.size(); ++m_Counter )
				{
					if ( !strncmp( matname, m_materialList[m_Counter]->m_name, strlen( matname ) ) )
					{
						mpi->materialIndex = m_Counter;
						foundmaterial = true;
						break;
					}
				}
			}
			if( !foundmaterial )
			{
				delete mpi;
				mpi = NULL;
			}
		}

		// vertex
		else if ( strncmp(ac_line, "v ", 2 ) == 0 )
		{
			float fX, fY, fZ;
			if ( sscanf( ac_line, "v %f %f %f", &fX, &fY, &fZ ) )
			{
				m_vertices.push_back( Vector3D( fX, fY, fZ ) );
			}
		}
		// vertex normals
		else if ( strncmp( ac_line, "vn ", 3 ) == 0 )
		{
			float fN1, fN2, fN3;
			if ( sscanf( ac_line, "vn %f %f %f", &fN1, &fN2, &fN3 ) )
			{
				m_normals.push_back( Vector3D( fN1, fN2, fN3 ) );
			}
		}
		// vertex normals
		else if ( strncmp( ac_line, "vt ", 3 ) == 0 )
		{
			float fT1 = 0, fT2 = 0;
			if ( sscanf( ac_line, "vt %f %f", &fT1, &fT2) )
			{
				m_texCoords.push_back( Vector3D( fT1, 1.0f-fT2, 0.0f ) );

			}
		}
		// face
		else if ( strncmp( ac_line, "f ", 2) == 0)
		{
			int	i3VI[4] = {0, 0, 0, -1};
			int	i3NI[4] = {0, 0, 0, -1};
			int i3TI[4] = {0, 0, 0, -1};

			//Quadrangel
			if ( 12 == sscanf( ac_line, "f %d//%d//%d %d//%d//%d %d//%d//%d %d//%d//%d", &i3VI[0], &i3TI[0], &i3NI[0], 
				&i3VI[1], &i3TI[1], &i3NI[1], 
				&i3VI[2], &i3TI[2], &i3NI[2],
				&i3VI[3], &i3TI[3], &i3NI[3] ) )
			{								
			}
			//Quadrangel
			else if ( 12 == sscanf( ac_line, "f %d/%d/%d %d/%d/%d %d/%d/%d %d/%d/%d", &i3VI[0], &i3TI[0], &i3NI[0], 
				&i3VI[1], &i3TI[1], &i3NI[1], 
				&i3VI[2], &i3TI[2], &i3NI[2],
				&i3VI[3], &i3TI[3], &i3NI[3] ) )
			{								
			}
			else if ( 9 == sscanf( ac_line, "f %d//%d//%d %d//%d/%d %d//%d//%d", &i3VI[0], &i3TI[0], &i3NI[0], 
				&i3VI[1], &i3TI[1], &i3NI[1], 
				&i3VI[2], &i3TI[2], &i3NI[2] ) )
			{								
			}
			else if ( 9 == sscanf( ac_line, "f %d/%d/%d %d/%d/%d %d/%d/%d", &i3VI[0], &i3TI[0], &i3NI[0], 
				&i3VI[1], &i3TI[1], &i3NI[1], 
				&i3VI[2], &i3TI[2], &i3NI[2] ) )
			{								
			}
			//Quadrangel
			else if ( 8 == sscanf( ac_line, "f %d//%d %d//%d %d//%d %d//%d", &i3VI[0], &i3NI[0], 
				&i3VI[1], &i3NI[1], 
				&i3VI[2], &i3NI[2],
				&i3VI[3], &i3NI[3] ) )
			{						
				++ui_normals_created;		
			}
			//Quadrangel
			else if ( 8 == sscanf( ac_line, "f %d/%d %d/%d %d/%d %d/%d", &i3VI[0], &i3NI[0], 
				&i3VI[1], &i3NI[1], 
				&i3VI[2], &i3NI[2],
				&i3VI[3], &i3NI[3] ) )
			{								
				++ui_normals_created;
			}
			else if ( 6 == sscanf( ac_line, "f %d//%d %d//%d %d//%d", &i3VI[0], &i3NI[0], 
				&i3VI[1], &i3NI[1], 
				&i3VI[2], &i3NI[2] ) )
			{
				++ui_normals_created;
			}
			else if ( 6 == sscanf( ac_line, "f %d/%d %d/%d %d/%d", &i3VI[0], &i3NI[0], 
				&i3VI[1], &i3NI[1], 
				&i3VI[2], &i3NI[2] ) )
			{
				++ui_normals_created;
			}
			//Quadrangel
			else if ( 4 == sscanf( ac_line, "f %d %d %d %d", &i3VI[0], &i3VI[1], &i3VI[2], &i3VI[3] ) )
			{
			}
			else if ( 3 == sscanf( ac_line, "f %d %d %d", &i3VI[0], &i3VI[1], &i3VI[2] ) )
			{
			}
			//obj-indices are 1-based
			i3VI[0] -= 1; i3VI[1] -= 1; i3VI[2] -= 1; i3VI[3] -= 1;
			i3NI[0] -= 1; i3NI[1] -= 1; i3NI[2] -= 1; i3NI[3] -= 1;
			i3TI[0] -= 1; i3TI[1] -= 1; i3TI[2] -= 1; i3TI[3] -= 1;
			m_triangles.push_back( Triangle( i3VI, i3TI, i3NI ) );

			//Quad-Case: Make second Triangle
			if ( i3VI[3] >= 0 )
			{
				int	i3V[3],	i3N[3], i3T[3];
				i3V[0] = i3VI[0]; i3V[1] = i3VI[2]; i3V[2] = i3VI[3];
				i3N[0] = i3NI[0]; i3N[1] = i3NI[2]; i3N[2] = i3NI[3];
				i3T[0] = i3TI[0]; i3T[1] = i3TI[2]; i3T[2] = i3TI[3];
				m_triangles.push_back( Triangle( i3V, i3T, i3N) );
				actFace++;
			}
			actFace++;
		}
	}

	if ( mpi != NULL )
	{
		mpi->toFace = actFace;
		m_meshPartInfo.push_back(mpi);
		m_objTexAvail = true;
	} 
	else 
	{ // no materials available
		m_meshPartInfo.push_back(new MeshPartInfo(0,actFace)); // set one materialnode with invalid index
		m_objTexAvail = false;
	}

	istr2_in.close();

	std::cout << "vertices:        " << m_vertices.size() << std::endl;
	std::cout << "normals:         " << m_normals.size() << std::endl;
	std::cout << "texcoords:       " << m_texCoords.size() << std::endl;
	std::cout << "normals created: " << ui_normals_created << std::endl;
	std::cout << "faces:           " << m_triangles.size() << std::endl;

	computeBoundingBox();
}

void Mesh::computeBoundingBox()
{
	// compute bounding box
	m_bBoxMin.x = m_bBoxMin.y = m_bBoxMin.z = FLT_MAX;
	m_bBoxMax.x = m_bBoxMax.y = m_bBoxMax.z = -FLT_MAX;
	for ( unsigned int i = 0; i < m_vertices.size(); ++i )
	{
		if(m_vertices[i].x < m_bBoxMin.x)
			m_bBoxMin.x = m_vertices[i].x;
		if(m_vertices[i].y < m_bBoxMin.y)
			m_bBoxMin.y = m_vertices[i].y;

		if(m_vertices[i].z < m_bBoxMin.z)
			m_bBoxMin.z = m_vertices[i].z;

		if(m_vertices[i].x > m_bBoxMax.x)
			m_bBoxMax.x = m_vertices[i].x;
		if(m_vertices[i].y > m_bBoxMax.y)
			m_bBoxMax.y = m_vertices[i].y;
		if(m_vertices[i].z > m_bBoxMax.z)
			m_bBoxMax.z = m_vertices[i].z;
	}
}

void Mesh::replaceFilename(char *tgt, const char *src, const char *newName)
{
	strcpy(tgt,src);
	char *p = strrchr(tgt,'/');
	if ( NULL == p)
	{
		p = strrchr(tgt,'\\');
		if( NULL == p)
		{
			strcpy(tgt,newName);
		}
	}
	p++;
	strcpy(p,newName);
}

void Mesh::loadTextureImages()
{
	for(unsigned int tc = 0; tc < m_textures.size(); tc++)
	{
		if(NULL == m_textures[tc].qImageTexture)
		{
			m_textures[tc].qImageTexture = new QImage;
			loadTexture(*(m_textures[tc].qImageTexture),m_textures[tc].textureId,m_textures[tc].texname);
		}
	}
}

void Mesh::assignTexture(char *filename)
{
	if ( !m_objTexAvail ) // no textures already available from obj file
	{
		m_meshPartInfo[0]->materialIndex = 0;
		m_materialList.push_back(new MaterialNode(0));
		m_textures.push_back(TextureNode());
		m_textures[0].qImageTexture = new QImage;
		strcpy(m_textures[0].texname,filename);
		loadTexture( *(m_textures[0].qImageTexture), m_textures[0].textureId, m_textures[0].texname );
	}
}


bool Mesh::loadTexture(QImage &qimg, GLuint &textureId, const char *filename) 
{
	QImage img;
	if( !img.load(QString(filename)) ) 
	{
		std::cerr << "Could not load image: " << filename << std::endl;
		return false;
	}
	QImage img2 = img.convertToFormat(QImage::Format_RGB32);
	qimg = img2.rgbSwapped();
	glGenTextures(1, &textureId);
	glBindTexture(GL_TEXTURE_2D, textureId);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	glTexImage2D(GL_TEXTURE_2D,
		0,                      //level
		GL_RGB8,                //internal format
		qimg.width(),			//width
		qimg.height(),			//height
		0,                      //border
		GL_RGBA,                //format
		GL_UNSIGNED_BYTE,       //type
		qimg.bits());			// pixel data

	return true;
}


// **********************
// ********* TODO *******
// **********************
// generate texture coordinates for all vertices
// store texture coordinates in  textcoords
// look in this->draw() for access to datastructure


void Mesh::generateSphereTexCoords()
{
	if( !m_objTexAvail )
	{
		float multp[2] = { 10.0, 15.0f }; // gehoert zu aufgabenteil b (multp[x],multp[y]) mal textur darstellen (GL_REPEAT)

		//correctTheta(multp[0]);
	}
}


void Mesh::generateCylinderTexCoords()
{
	if( !m_objTexAvail )
	{
		float multp[2] = { 5.0, 10.0f }; // gehoert zu aufgabenteil b (multp[x],multp[y]) mal textur darstellen (GL_REPEAT)

		//correctTheta(multp[0]);
	}
}

void Mesh::correctTheta(float multp)
{
	Vector3D center( (m_bBoxMin+m_bBoxMax) / 2 );

	// update triangle texture indices and
	// correct border cases, faces crossing the 0/1 texcoord border
	unsigned int texcoordCount = static_cast<unsigned int>(m_texCoords.size());
	for(unsigned int t = 0; t < m_triangles.size(); t++)
	{
		Triangle &triangle = m_triangles[t];
		bool backVert = false, frontVert = false, leftVert = false;
		for ( int j = 0; j < 3; j++)
		{
			Vector3D &vertex = m_vertices[triangle.v[j]];
			Vector3D delta = vertex - center;
			if(delta.z < 0) backVert = true;
			if(delta.z >= 0) frontVert = true;
			if(delta.x < 0) leftVert = true;
		}
		if( backVert && frontVert && leftVert )
		{
			for ( int j = 0; j < 3; j++)
			{
				Vector3D &vertex = m_vertices[triangle.v[j]];
				Vector3D delta = vertex - center;
				if( delta.z < 0 ) 
				{
					Vector3D newCoord(m_texCoords[triangle.t[j]]);
					newCoord.x += multp;
					m_texCoords.push_back(newCoord);
					triangle.t[j] = texcoordCount;
					texcoordCount++;
				}
			}
		}
	}
}
